/*
DO NOT EDIT!

This file is autogenerated by synphot/setup_package.py.  To edit
its contents, edit synphot/docstrings.py

The weirdness here with strncpy is because some C compilers, notably
MSVC, do not support string literals greater than 256 characters.
*/

#include <string.h>
#include "docstrings.h"

#if defined(_MSC_VER)
char doc_calcbinflux[594];

void fill_docstrings(void)
{
   strncpy(doc_calcbinflux + 0, "calcbinflux(len_binwave, i_beg, i_end, avflux, deltaw)\n\nSum over each bin.\n\nParameters\n----------\nlen_binwave : int\n    Number of wavelength bin centers.\n\ni_beg, i_end : array-like\n    Locations of bin edges in ``deltaw``.\n\navflux : array-like\n    Average ", 256);
   strncpy(doc_calcbinflux + 256, "flux associated with ``deltaw``.\n\ndeltaw : array-like\n    Delta of merge wavelengths (native + centers + edges).\n    Values are in ascending order.\n\nReturns\n-------\nbinflux : array-like\n    Integrated flux associated with given bins in ascending order.\n\nin", 256);
   strncpy(doc_calcbinflux + 512, "twave : array-like\n    Integrated delta wavelength associated with ``binflux``.\n\n\0", 82);


}

#else /* UNIX */
char doc_calcbinflux[594] = "calcbinflux(len_binwave, i_beg, i_end, avflux, deltaw)\n\nSum over each bin.\n\nParameters\n----------\nlen_binwave : int\n    Number of wavelength bin centers.\n\ni_beg, i_end : array-like\n    Locations of bin edges in ``deltaw``.\n\navflux : array-like\n    Average flux associated with ``deltaw``.\n\ndeltaw : array-like\n    Delta of merge wavelengths (native + centers + edges).\n    Values are in ascending order.\n\nReturns\n-------\nbinflux : array-like\n    Integrated flux associated with given bins in ascending order.\n\nintwave : array-like\n    Integrated delta wavelength associated with ``binflux``.\n\n\0";

#endif
